MODEL_CONFIG = {
    "models/gemini-1.5-pro": {
        "context_tokens": 1048576,
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.0000035,
        "completion_cost_per_token": 0.0000105,
    },
    "models/gemini-1.5-flash": {
        "context_tokens": 1048576,
        "max_tokens": 8192,
        "prompt_cost_per_token": 0.00000035,
        "completion_cost_per_token": 0.00000105,
    },
}
